/*
 *	change and/or display date
 *
 *	date [ -u ] [ [ [ mm ] dd ] hhmm [ yy ] ]         [ +format ] (sysV)
 *	date [ -u ] [ [ [ [ yy ] mm ] dd ] hhmm [ .ss ] ] [ +format ] (bsd)
 *
 *	format:		result:
 *	%%		%
 *	%n		\n
 *	%t		\t
 *	%m		month - 01 to 12
 *	%d		day of month - 01 to 31
 *	%y		year mod 100 - 00 to 99
 * #ifdef EXTRAS
 *	%Y		year - 1970 to 2069			(special)
 * #endif
 *	%D		date as mm/dd/yy
 *	%H		hour - 00 to 23
 *	%M		minute - 00 to 59
 *	%S		second - 00 to 59
 *	%T		time as hh:mm:ss
 *	%g		timezone as GMT+/-nn:nn			(special)
 *	%j		julian date - 001 to 366
 *	%w		day of week - sunday = 0
 *	%a		abbreviated weekday - Sun to Sat
 *	%h		abbreviated month - Jan to Dec
 *	%r		time as hh:mm:ss AM/PM
 *	%z		abbreviated time zone name		(special)
 *	%G		use GMT from now on			(special)
 *	%L		use local time from now on		(special)
 */

#define EXTRAS		/* include code for 'Y' format */
/* #define USEASCTIME	/* use asctime(3) for 'a' and 'h' format */
/* #define WANTBUGS	/* include diab bugs */

#include <ctype.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <time.h>
#include <utmp.h>
#include <bool.h>

#define BUFSZ 200
#define MAXTM 30

/*
** global variables
*/

char *program;
struct timeb tp = { 0 };	/* time info structure			*/
struct tm timegmt = { 0 };	/* holder for result of gmtime ()	*/
struct tm timeloc = { 0 };	/* holder for result of localtime ()	*/
struct tm *mtime;		/* pointer to selected timestruc.	*/
int monsize[12] = {
	31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};
#ifndef USEASCTIME
char *monnames[12] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};
char *daynames[7] = {
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};
#endif
struct utmp wtmp_bef = {
	"|",
	"",
	0
};
struct utmp wtmp_aft = {
	"{",
	"",
	0
};
#ifdef EXTRAS
 char format[] = "+%a %h %d %T %z %Y";		/* default format */
#else
 char format[] = "+%a %h %d %T %z 19%y";	/* default format */
#endif

/*
** date
*/

main ( argc, argv )
int argc;
char *argv[];

{
	int year = -1;
	int mon = -1;
	int day = -1;
	int hour = -1;
	int min = -1;
	int sec = 0;
	int i;
	int filedsc;
	int result = 0;
	int pos = 0;
	int oldver = 0;
	int uflg = FALSE;
#ifdef USEASCTIME
	char timestr[MAXTM];
#endif
	char buf[BUFSZ];
	char *a;
	char *b;
	time_t timecnt;

	/*
	** get program name
	*/

	program = *argv;

	--argc;
	++argv;

	/*
	** look for -u switch
	*/

	if ( argc && *argv && !strcmp ( *argv, "-u" ) ) {
		--argc;
		++argv;
		uflg = TRUE;
	}

	ftime ( &tp );
	if ( uflg ) {
		mtime = gmtime ( &tp.time );
	} else {
		mtime = localtime ( &tp.time );
	}
#ifdef USEASCTIME
	strncpy ( timestr, asctime ( mtime ), MAXTM );
#endif

	/*
	** if we have an argument, but the argument does not start
	** with a + sign, then we should set the time
	*/

	/*
	** warning: very twisted logic ahead, watch out for magic numbers
	*/

	if ( *argv && **argv != '+' ) {
		for ( i = 0; i < strlen ( *argv ); i++) {
			buf[i] = argv[0][i];
			if ( ( i==4||i==6||i==8||i==10 ) && buf[i] == '.' ) {
				if ( oldver ) {
					result++;
				} else {
					oldver = i;
				}
			} else {
				if ( !isdigit ( buf[i] ) ) {
					result++;
				}
			}
		}
		buf[i] = '\0';

		if ( ( oldver && oldver != (i-3) )
		  || result
		  || ( i != 4 && (i < 6 || i > 11) && i != 13 ) ) {
			fprintf ( stderr, "%s: bad conversion\n", program );
			exit ( 1 );
		}

		/*
		** get second
		*/

		if ( i & 1 ) {
			sec = atoi ( &buf[i - 2] );
			i -= 3;
			buf[i] = '\0';
		}

		/*
		** get year
		*/

		if ( i == 10 ) {
			if ( oldver || *buf > '1') {	/* move year */
				buf[10] = buf[0];
				buf[11] = buf[1];
				buf[12] = '\0';
				for ( a = buf; ( *a = a[2] ); a++ ) {
					/*VOID*/
				}
			}
			year = atoi ( &buf[8] );
			i -= 2;
		} else {
			year = mtime->tm_year;
		}

		/*
		** get minute
		*/

		buf[8] = '\0';
		min = atoi ( &buf[i - 2] );

		/*
		** get hour
		*/

		buf[i - 2] = '\0';
		hour = atoi ( &buf[i - 4] );

		/*
		** get day
		*/

		buf[i - 4] = '\0';
		if ( i >= 6 ) {
			day = atoi ( &buf[i - 6] );
			buf[i - 6] = '\0';
		} else {
			day = mtime->tm_mday;
		}

		/*
		** get month
		*/

		if ( i >= 8 ) {
			mon = atoi ( buf );
		} else {
			mon = mtime->tm_mon;
		}

		/*
		** fix hour
		*/

		if ( hour == 24 ) {
			hour = 0;
#ifndef WANTBUGS
			day++;		/* system V does this */
#endif
		}

		/*
		** validate input
		*/

		if ( mon  < 1
		  || mon  > 12
		  || day  < 1
		  || day  > monsize[mon-1]
		  || ( day == 29 && mon == 2 && dysize ( year ) == 365 )
		  || hour < 0
		  || hour > 23
		  || min  < 0
		  || min  > 59
		  || sec  < 0
		  || sec  > 59 ) {
			fprintf ( stderr, "%s: bad conversion\n", program );
			exit ( 1 );
		}

		/*
		** calculate number of seconds since Jan 1, 1970
		*/

		/*
		** first, start with zero time
		*/

		timecnt = 0;

		/*
		** add current century to the year
		*/

		year += 1900;

		/*
		** handle the years between 2000 and 2069
		*/

		if ( year < 1970 ) {
			year += 100;	/* 20nn */
		}

		/*
		** compensate if the current year is not a leap year
		*/

		if ( dysize ( year ) == 365 && mon > 2 ) {
			timecnt--;
		}

		/*
		** how many days do all year add up to
		*/

		while ( --year > 1969 ) {
			timecnt += dysize ( year );
		}

		/*
		** same for the months
		*/

		while ( --mon ) {
			timecnt += monsize[mon - 1];
		}

		/*
		** and the days in the current month
		*/

		timecnt += day - 1;

		/*
		** convert days to hours and add current hour
		*/

		timecnt = 24 * timecnt + hour;

		/*
		** convert hours to minutes and add current minute
		*/

		timecnt = 60 * timecnt + min;

		/*
		** if not gmt, then add timezone and maybe daylight savings
		*/

		if ( !uflg ) {
			timecnt += tp.timezone;
			if ( tp.dstflag ) {
				timecnt -= 60;
			}
		}

		/*
		** convert minutes to seconds and add current second
		*/

		timecnt = 60 * timecnt + sec;

#ifdef DEBUG
		fprintf (
		  stderr,
		  "%s: y:%d m:%d d:%d h:%d m:%d s:%d t:%ld\n",
		  program, year, mon, day, hour, min, sec, timecnt
		);
#endif

		/*
		** get time count before change of date
		*/

		time ( &wtmp_bef.ut_time );

		/*
		** ty to change the date
		*/

		if ( stime ( &timecnt ) == -1 ) {
			fprintf ( stderr, "%s: no permission\n", program );
			exit ( 1 );
		}

		if ( ( filedsc = open ( "/usr/adm/wtmp", 1 ) ) == -1 ) {
/**
			fprintf (
			  stderr,
			  "%s: cannot open file %s\n",
			  program, "/usr/adm/wtmp"
			);
			result++;
**/
		} else {
			/*
			** get time count after change
			*/

			time ( &wtmp_aft.ut_time );

			/*
			** append to file
			*/

			lseek ( filedsc, 0L, 2 );

			/*
			** try to write the two records
			*/

			if ( write ( filedsc, &wtmp_bef, sizeof (struct utmp) ) != sizeof (struct utmp)
			  || write ( filedsc, &wtmp_aft, sizeof (struct utmp) ) != sizeof (struct utmp) ) {
				fprintf (
				  stderr,
				  "%s: write error %s\n",
				  program, "/usr/adm/wtmp"
				);
				result++;
			}
			close ( filedsc );
		}
		argv++;
		argc--;
	}

	/*
	** warning: twisted logic ahead
	**
	** accept the following usages:
	**   date datespec           ==> exit now
	**   date datespec +format   ==> display date
	**   date datespec anything  ==> usage info and exit
	**   date +format            ==> display date
	**   date                    ==> display date
	*/

	if ( *argv ) {
		if ( ( **argv != '+' ) || argv[1] ) {	/* unknown */
			fprintf (
			  stderr,
#ifdef WANTBUGS
			  "usage: %s [mmddhhmm[yy]] [+format]",
#else
			  "usage: %s [-u] [[[[yy]mm]dd]hhmm[.ss]]|[[[mm]dd]hhmm[yy]] [+format]",
#endif
			  program
			);
			exit ( 1 );
		}
	} else {
		if ( wtmp_bef.ut_time ) {		/* date is set */
			exit ( result );		/* no arg's left */
		} else {
			*argv = format;
		}
	}

	ftime ( &tp );

	mtime = localtime ( &tp.time );
#ifdef WANTBUGS
	for ( a = &timeloc, b = mtime, i = 0;
#else
	for ( a = (char *)&timeloc, b = (char *)mtime, i = 0;
#endif
	  i++ < sizeof (struct tm);
	  *a++ = *b++ ) {
		/*VOID*/
	}

	mtime = gmtime ( &tp.time );
#ifdef WANTBUGS
	for ( a = &timegmt, b = mtime, i = 0;
#else
	for ( a = (char *)&timegmt, b = (char *)mtime, i = 0;
#endif
	  i++ < sizeof (struct tm);
	  *a++ = *b++ ) {
		/*VOID*/
	}

	if ( uflg ) {
		mtime = &timegmt;
	} else {
		mtime = &timeloc;
	}
#ifdef USEASCTIME
# ifdef WANTBUGS
	strncpy ( timestr, asctime ( &timegmt ), MAXTM );
# else
	strncpy ( timestr, asctime ( mtime ), MAXTM );
# endif
#endif

	while( *++*argv && pos < BUFSZ - 13 ) {
		if ( **argv != '%' ) {
			buf[pos++] = **argv;
		} else {
			switch ( *++*argv ) {
			case '%':	/* Doubled			*/
				buf[pos++] = '%';
				break;
			case 'n':	/* Insert a newline		*/
				buf[pos++] = '\n';
				break;
			case 't':	/* Insert a tab			*/
				buf[pos++] = '\t';
				break;
			case 'm':	/* Month - 01 to 12		*/
				sprintf ( &buf[pos], "%02d", mtime->tm_mon+1 );
				pos += 2;
				break;
			case 'd':	/* Day of month - 01 to 31	*/
				sprintf ( &buf[pos], "%02d", mtime->tm_mday );
				pos += 2;
				break;
			case 'y':	/* Year mod 100			*/
				sprintf ( &buf[pos], "%02d", mtime->tm_year );
				pos += 2;
				break;
#ifdef EXTRAS
			case 'Y':	/* Year				*/
				i = mtime->tm_year + 1900;
				if ( i < 1970 ) {
					i += 100;	/* 20nn */
				}
				sprintf ( &buf[pos], "%04d", i );
				pos += 4;
				break;
#endif
			case 'D':	/* Date as mm/dd/yy		*/
				sprintf (
				  &buf[pos], "%02d/%02d/%02d",
				  mtime->tm_mon+1,
				  mtime->tm_mday,
				  mtime->tm_year
				);
				pos += 8;
				break;
			case 'H':	/* Hour - 00 to 23		*/
				sprintf ( &buf[pos], "%02d", mtime->tm_hour );
				pos += 2;
				break;
			case 'M':	/* Minute - 00 to 59		*/
				sprintf ( &buf[pos], "%02d", mtime->tm_min );
				pos += 2;
				break;
			case 'S':	/* Second - 00 to 59		*/
				sprintf ( &buf[pos], "%02d", mtime->tm_sec );
				pos += 2;
				break;
			case 'T':	/* Time as hh:mm:ss		*/
				sprintf (
				  &buf[pos], "%02d:%02d:%02d",
				  mtime->tm_hour,
				  mtime->tm_min,
				  mtime->tm_sec
				);
				pos += 8;
				break;
			case 'g':	/* Timezone - GMT+/-nn:nn	*/
				sprintf (
				  &buf[pos], "GMT%c%02d:%02d",
				  (tp.timezone >= 0 ? '-' : '+'),
				  abs ( tp.timezone ) / 60,
				  abs ( tp.timezone ) % 60
				);
				pos += 9;
				break;
			case 'j':	/* Julian date - 001 to 366	*/
				sprintf ( &buf[pos], "%03d", mtime->tm_yday+1 );
				pos += 3;
				break;
			case 'w':	/* Day of week - Sunday = 0	*/
				sprintf ( &buf[pos], "%01d", mtime->tm_wday );
				pos++;
				break;
			case 'a':	/* Abbreviated weekday - Sun to Sat*/
#ifdef USEASCTIME
				sprintf ( &buf[pos], "%.3s", timestr );
#else
				sprintf ( &buf[pos], "%.3s", daynames[mtime->tm_wday] );
#endif
				pos += 3;
				break;
			case 'h':	/* Abbreviated month - Jan to Dec*/
#ifdef USEASCTIME
				sprintf ( &buf[pos], "%.3s", &timestr[4] );
#else
				sprintf ( &buf[pos], "%.3s", monnames[mtime->tm_mon] );
#endif
				pos += 3;
				break;
			case 'r':	/* Time in AM/PM notation	*/
				i = mtime->tm_hour % 12;
				if ( i == 0 ) {
					i = 12;
				}
				sprintf (
				  &buf[pos], "%02d:%02d:%02d %cM",
				  i,
				  mtime->tm_min,
				  mtime->tm_sec,
				  (mtime->tm_hour >= 12 ? 'P' : 'A')
				);
				pos += 11;
				break;
			case 'z':	/* Time zone name		*/
				if ( ( a = tzname[timeloc.tm_isdst] ) ) {
					sprintf ( &buf[pos], "%s", a );
					pos += strlen ( a );
				}
				break;
			case 'G':	/* Time measured in GMT		*/
				mtime = &timegmt;
#ifdef USEASCTIME
# ifndef WANTBUGS
				strncpy ( timestr, asctime ( mtime ), MAXTM );
# endif
#endif
				break;
			case 'L':	/* Time measured in local time	*/
				mtime = &timeloc;
#ifdef USEASCTIME
# ifndef WANTBUGS
				strncpy ( timestr, asctime ( mtime ), MAXTM );
# endif
#endif
				break;
			default:
				fprintf (
				  stderr,
				  "%s: bad format character %s\n",
				  program,
				  *argv
				);
				exit ( 1 );
			}
		}
	}

	buf[pos++] = '\n';
	buf[pos]   = '\0';
	printf ( "%s", buf );

	exit ( result );
}
